@@ -1,8 +1,10 @@
 from __future__ import annotations
 
 from collections import Mapping  # pylint: disable=no-name-in-module
+from dataclasses import is_dataclass
 from itertools import chain
-from typing import Optional, List, Dict
+from typing import Optional, List, Dict, Union
+from ._util import elements
 
 
 class WildcardMatch:
@@ -33,10 +35,7 @@ class MatchContext:
             self.groups[key] = value
 
     def __getitem__(self, item):
-        values = self.groups[item]
-        if self.multimatch and len(values) == 1:
-            return values[0]
-        return values
+        return self.groups[item]
 
     def __contains__(self, item):
         return item in self.groups
@@ -55,14 +54,32 @@ class MatchContext:
         if isinstance(pattern, Pattern):
             return pattern.match(value, ctx=self, strict=strict)
 
+        if is_dataclass(value):
+            if is_dataclass(pattern):
+                pattern_type = type(pattern)
+                pattern_dict = pattern.__dict__
+            elif isinstance(pattern, Dataclass):
+                pattern_type = pattern.type
+                pattern_dict = pattern.dict
+            else:
+                return self.no_match()
+
+            if not issubclass(type(value), pattern_type):
+                return self.no_match()
+            return _match_mapping(value.__dict__, pattern_dict, ctx=self, strict=strict)
+
         if isinstance(pattern, dict):
-            return _match_dict(value, pattern, ctx=self, strict=strict)
+            return _match_mapping(value, pattern, ctx=self, strict=strict)
 
         if isinstance(pattern, tuple):
-            return _match_tuple(value, pattern, ctx=self, strict=strict)
+            if not isinstance(value, tuple) or strict and type(value) != tuple:
+                return self.no_match()
+            return _match_sequence(value, pattern, ctx=self)
 
         if isinstance(pattern, list):
-            return _match_list(value, pattern, ctx=self, strict=strict)
+            if strict and type(value) != list:
+                return self.no_match()
+            return _match_sequence(value, pattern, ctx=self)
 
         return self.no_match()
 
@@ -81,15 +98,19 @@ class MatchContext:
             self.wildcards[id_] = WildcardMatch(len(self.wildcards))
         self.wildcards[id_].set(value)
 
+    def get_wildcard_matches(self) -> List:
+        wildcard_matches: List = [None] * len(self.wildcards)
+
+        for wildcard_match in self.wildcards.values():
+            wildcard_matches[wildcard_match.index] = wildcard_match.get()
+        return wildcard_matches
+
 
 class MatchResult(Mapping):
     def __init__(self, *, matches: bool, context: MatchContext):
         self._matches: bool = matches
         self._context: MatchContext = context
-        self._wildcard_matches: List = [None] * len(context.wildcards)
-
-        for wildcard_match in context.wildcards.values():
-            self._wildcard_matches[wildcard_match.index] = wildcard_match.value
+        self._wildcard_matches: List = context.get_wildcard_matches()
 
     def __bool__(self):
         return self._matches
@@ -97,6 +118,9 @@ class MatchResult(Mapping):
     def __getitem__(self, item):
         return self._context[item]
 
+    def __getattr__(self, item):
+        return self._context[item]
+
     def __contains__(self, item):
         return item in self._context
 
@@ -118,6 +142,11 @@ class MatchResult(Mapping):
     def wildcard_matches(self) -> List:
         return self._wildcard_matches
 
+    def bind(self, target) -> MatchResult:
+        for k, v in self.items():
+            target[k] = v
+        return self
+
 
 class Capturable:
     def __rshift__(self, other):
@@ -127,6 +156,44 @@ class Capturable:
         return Capture(self, name=other)
 
 
+class Nested:
+    def descend(self, f):
+        raise NotImplementedError
+
+
+class Dataclass(Nested):
+    def __init__(self, type_: type, dict_: dict):
+        self.type = type_
+        self.dict = dict_
+
+    def descend(self, f):
+        dict_ = {}
+        for k, v in self.dict.items():
+            dict_[k] = f(v)
+        return Dataclass(self.type, dict_)
+
+
+def transform(pattern, f):
+    def rf(p):
+        return transform(p, f)
+
+    if is_dataclass(pattern):
+        pattern = Dataclass(type(pattern), pattern.__dict__)
+
+    if isinstance(pattern, Nested):
+        return f(pattern.descend(rf))
+    if isinstance(pattern, dict):
+        kvs = {}
+        for k, v in pattern.items():
+            kvs[rf(k)] = rf(v)
+        return f(kvs)
+    if isinstance(pattern, tuple):
+        return f(tuple(rf(i) for i in pattern))
+    if isinstance(pattern, list):
+        return f(list(rf(i) for i in pattern))
+    return f(pattern)
+
+
 class Pattern(Capturable):
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
         raise NotImplementedError
@@ -143,6 +210,16 @@ class Pattern(Capturable):
     def __invert__(self):
         return Not(self)
 
+    def __eq__(self, other):
+        return type(self) == type(other) and self.__dict__ == other.__dict__
+
+    def __hash__(self):
+        return hash(tuple(elements(self)))
+
+    def __repr__(self):
+        s = tuple(f"{k}={repr(v)}" for k, v in self.__dict__.items())
+        return f"{type(self).__name__}({', '.join(s)})"
+
 
 class StringPattern:
     """Experimental"""
@@ -151,7 +228,7 @@ class StringPattern:
         raise NotImplementedError
 
 
-class String(Pattern):
+class String(Pattern, Nested):
     """Experimental"""
 
     def __init__(self, *patterns):
@@ -169,7 +246,8 @@ class String(Pattern):
                     ctx[name] = pattern
                 return pattern
         elif isinstance(pattern, StringPattern):
-            if (result := pattern.string_match(remaining, ctx=ctx)) is not None:
+            result = pattern.string_match(remaining, ctx=ctx)
+            if result is not None:
                 if name:
                     ctx[name] = result
                 return result
@@ -178,24 +256,32 @@ class String(Pattern):
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
         remaining = value
         for p in self._patterns:
-            if (matched := self.match_pattern(remaining=remaining, pattern=p, ctx=ctx)) is not None:
+            matched = self.match_pattern(remaining=remaining, pattern=p, ctx=ctx)
+            if matched is not None:
                 remaining = remaining[len(matched):]
             else:
                 return ctx.no_match()
         return ctx.match_if(not remaining)
 
+    def descend(self, f):
+        return String(*(f(p) for p in self._patterns))
 
-class Capture(Pattern):
-    def __init__(self, pattern, /, *, name):
+
+class Capture(Pattern, Nested):
+    def __init__(self, pattern, *, name):
         self._pattern = pattern
         self._name = name
 
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
-        if result := ctx.match(value, self._pattern):
+        result = ctx.match(value, self._pattern)
+        if result:
             ctx[self._name] = value
             return result
         return ctx.no_match()
 
+    def descend(self, f):
+        return Capture(pattern=f(self._pattern), name=self._name)
+
     @property
     def pattern(self):
         return self._pattern
@@ -205,8 +291,8 @@ class Capture(Pattern):
         return self._name
 
 
-class Some(Capturable):
-    def __init__(self, pattern, /, *,
+class Some(Capturable, Nested):
+    def __init__(self, pattern, *,
                  at_least: Optional[int] = None,
                  at_most: Optional[int] = None,
                  exactly: Optional[int] = None):
@@ -235,62 +321,74 @@ class Some(Capturable):
             return count >= self.at_least
         return True
 
+    def descend(self, f):
+        return Some(pattern=f(self.pattern), at_least=self.at_least, at_most=self.at_most)
+
 
 class Remaining(Some):
     pass
 
 
-class Strict(Pattern):
-    def __init__(self, pattern, /):
+class Strict(Pattern, Nested):
+    def __init__(self, pattern):
         self._pattern = pattern
 
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
         return ctx.match(value, self._pattern, strict=True)
 
+    def descend(self, f):
+        return Strict(pattern=f(self._pattern))
+
 
 class Value(Pattern):
-    def __init__(self, value, /):
+    def __init__(self, value):
         self._value = value
 
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
-        if self._value is Ellipsis:
-            ctx.record(self, value)
-            return ctx.matches()
         if strict:
             if type(self._value) != type(value):
                 return ctx.no_match()
         return ctx.match_if(self._value == value)
 
 
-class OneOf(Pattern, StringPattern):
+class OneOf(Pattern, StringPattern, Nested):
     def __init__(self, *patterns):
         self._patterns = patterns
 
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
         for pattern in self._patterns:
-            if result := ctx.match(value, pattern):
+            result = ctx.match(value, pattern)
+            if result:
                 return result
         return ctx.no_match()
 
     def string_match(self, remaining, *, ctx: MatchContext) -> Optional[str]:
         for p in self._patterns:
-            if (result := String.match_pattern(remaining=remaining, pattern=p, ctx=ctx)) is not None:
+            result = String.match_pattern(remaining=remaining, pattern=p, ctx=ctx)
+            if result is not None:
                 return result
         return None
 
+    def descend(self, f):
+        return OneOf(*(f(p) for p in self._patterns))
 
-class AllOf(Pattern):
+
+class AllOf(Pattern, Nested):
     def __init__(self, *patterns):
         self._patterns = patterns
 
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
         for pattern in self._patterns:
-            if not (result := ctx.match(value, pattern)):
+            result = ctx.match(value, pattern)
+            if not result:
                 return result
         return ctx.matches()
 
+    def descend(self, f):
+        return AllOf(*(f(p) for p in self._patterns))
 
-class Either(Pattern):
+
+class Either(Pattern, Nested):
     def __init__(self, left, right):
         self._left = left
         self._right = right
@@ -300,9 +398,12 @@ class Either(Pattern):
         match_right = ctx.match(value, self._right, strict=strict)
         return ctx.match_if(bool(match_left) != bool(match_right))
 
+    def descend(self, f):
+        return Either(f(self._left), f(self._right))
 
-class Not(Pattern):
-    def __init__(self, pattern, /):
+
+class Not(Pattern, Nested):
+    def __init__(self, pattern):
         self._pattern = pattern
 
     def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
@@ -310,41 +411,48 @@ class Not(Pattern):
             return ctx.no_match()
         return ctx.matches()
 
+    def descend(self, f):
+        return Not(pattern=f(self._pattern))
 
-def _match_dict(value, pattern, *, ctx: MatchContext, strict: bool) -> MatchResult:
+
+def _match_mapping(value, pattern: dict, *, ctx: MatchContext, strict: bool) -> MatchResult:
+    to_be_matched = {}
     try:
         items = value.items()
     except (AttributeError, TypeError):
         return ctx.no_match()
-    matched_keys = set()
-    for key, value in items:
-        if key in pattern:
-            if not (result := ctx.match(value, pattern[key])):
-                return result
-            matched_keys.add(key)
-        else:
-            if strict:
-                return ctx.no_match()
-    for key in pattern:
-        if key not in matched_keys:
+    for key, val in items:
+        to_be_matched[key] = val
+    patterns = []
+    for key, val_pattern in pattern.items():
+        if isinstance(key, Pattern):
+            patterns.append((key, val_pattern))
+            continue
+        try:
+            val = to_be_matched[key]
+        except KeyError:
             return ctx.no_match()
-    return ctx.matches()
-
-
-def _match_tuple(value, pattern, *, ctx: MatchContext, strict: bool) -> MatchResult:
-    if strict and type(value) != tuple:
-        return ctx.no_match()
-    if not isinstance(value, tuple) or len(pattern) != len(value):
-        return ctx.no_match()
-    for p, v in zip(pattern, value):
-        if not (result := ctx.match(v, p)):
+        result = ctx.match(val, val_pattern)
+        if not result:
             return result
-    return ctx.matches()
+        del to_be_matched[key]
+    possibly_mismatching_keys = set()
+    for key_pattern, val_pattern in patterns:
+        keys_to_remove = []
+        for key, val in to_be_matched.items():
+            if ctx.match(key, key_pattern):
+                if ctx.match(val, val_pattern):
+                    keys_to_remove.append(key)
+                    if key in possibly_mismatching_keys:
+                        possibly_mismatching_keys.remove(key)
+                elif not isinstance(key_pattern, Underscore):
+                    possibly_mismatching_keys.add(key)
+        for key in keys_to_remove:
+            del to_be_matched[key]
+    return ctx.match_if(not possibly_mismatching_keys and (not strict or not to_be_matched))
 
 
-def _match_list(value, pattern, *, ctx: MatchContext, strict: bool) -> MatchResult:
-    if strict and type(value) != list:
-        return ctx.no_match()
+def _match_sequence(value, pattern: Union[tuple, list], *, ctx: MatchContext) -> MatchResult:
     try:
         it = iter(value)
     except TypeError:
@@ -389,7 +497,8 @@ def _match_list(value, pattern, *, ctx: MatchContext, strict: bool) -> MatchResu
                 item = next(it)
         except StopIteration:
             return ctx.no_match()
-        if not (result := ctx.match(item, current_pattern)):
+        result = ctx.match(item, current_pattern)
+        if not result:
             return result
     try:
         next(it)
@@ -398,4 +507,10 @@ def _match_list(value, pattern, *, ctx: MatchContext, strict: bool) -> MatchResu
         return ctx.match_if(not item_queued)
 
 
-_ = Value(...)
+class Underscore(Pattern):
+    def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
+        ctx.record(self, value)
+        return ctx.matches()
+
+
+_ = Underscore()
