@@ -1,25 +1,18 @@
 import inspect
+import typing
 from typing import Callable, Dict, List
 
-import typing
-
-from apm.match import match
+from .error import MatchError
+from .match import match
 
 
 class Match:
     def __init__(self, pattern):
         self._pattern = pattern
 
-    @property
-    def pattern(self):
-        return self._pattern
-
-    def __call__(self):
-        return self
-
-
-class NoMatchError(Exception):
-    pass
+    def __call__(self, value):
+        # typing.get_type_hints requires annotations to be callable, otherwise it bombs out with a TypeError
+        return match(value, self._pattern)
 
 
 # noinspection PyDefaultArgument
@@ -38,13 +31,14 @@ def overload(fn: Callable, func_map: Dict[str, List[Callable]] = {}):
             except TypeError:
                 continue
             matches = True
+            # resolve the annotations which are strings as per __future__
             type_hints = typing.get_type_hints(func)
             for name, value in bound.arguments.items():
                 if name not in type_hints:
                     continue
                 annotation = type_hints[name]
                 if isinstance(annotation, Match):
-                    if not match(value, annotation.pattern):
+                    if not annotation(value):
                         matches = False
                         break
                 try:
@@ -57,7 +51,7 @@ def overload(fn: Callable, func_map: Dict[str, List[Callable]] = {}):
             if not matches:
                 continue
             return func(*bound.args, **bound.kwargs)
-        raise NoMatchError
+        raise MatchError(f"No match(args={repr(args)}, kwargs={repr(kwargs)})")
 
     return wrapper
 
