@@ -1,6 +1,7 @@
 from __future__ import annotations
 
 import unittest
+from dataclasses import dataclass
 from datetime import datetime
 
 from apm import *
@@ -36,9 +37,9 @@ class BasicUseCases(unittest.TestCase):
         self.assertFalse(match(1, Strict(1.0)))
 
     def test_truish(self):
-        self.assertTrue(match(1, Truish()))
-        self.assertTrue(match(2.9, Truish()))
-        self.assertFalse(match("", Truish()))
+        self.assertTrue(match(1, IsTruish))
+        self.assertTrue(match(2.9, IsTruish))
+        self.assertFalse(match("", IsTruish))
 
     def test_length(self):
         self.assertTrue(match([], Length(0)))
@@ -105,7 +106,7 @@ class BasicUseCases(unittest.TestCase):
             [1, 2, 3, Remaining(InstanceOf(int), at_least=2)]))
 
     def test_capture(self):
-        self.assertTrue(result := match(
+        result = match(
             {
                 "This": "Is A Rather Complex Beast",
                 "Created-At": "Sun Jan  3 04:08:57 CET 2021",
@@ -121,16 +122,18 @@ class BasicUseCases(unittest.TestCase):
                     "LastName": Capture(..., name="last_name"),
                 }, name="user")
             }
-        ))
+        )
+        self.assertTrue(result)
         self.assertEqual("Jane", result['first_name'])
         self.assertEqual("Doe", result['last_name'])
         self.assertEqual(102384, result['user']['UserId'])
 
     def test_remaining_with_capture(self):
-        self.assertTrue(result := match(
+        result = match(
             [1, 2, 3, 4],
             [1, 2, Capture(Remaining(...), name="tail")]
-        ))
+        )
+        self.assertTrue(result)
         self.assertEqual([3, 4], result['tail'])
 
     def test_and(self):
@@ -231,18 +234,19 @@ class BasicUseCases(unittest.TestCase):
             }
         }
 
-        self.assertTrue(result := match(record, At("foo.bar.quux", {"value": Capture(..., name="value")})))
+        result = match(record, At("foo.bar.quux", {"value": Capture(..., name="value")}))
+        self.assertTrue(result)
         self.assertEqual("deeply nested", result['value'])
 
-        self.assertTrue(result := match(record, At(['foo', 'bar', 'quux'], {"value": Capture(..., name="value")})))
+        result = match(record, At(['foo', 'bar', 'quux'], {"value": Capture(..., name="value")}))
+        self.assertTrue(result)
         self.assertEqual("deeply nested", result['value'])
 
         self.assertFalse(match(record, At("foo.bar.quux.baz", InstanceOf(int))))
 
     def test_string(self):
         path = "https://somehost/foo=1/bar=2/"
-
-        self.assertTrue(result := match(path, String(
+        result = match(path, String(
             Capture(OneOf("https", "http"), name="protocol"),
             "://",
             Capture(Regex("[a-zA-Z_-]+"), name="host"),
@@ -253,17 +257,17 @@ class BasicUseCases(unittest.TestCase):
             "bar=",
             Capture(Regex("[^=/]+"), name="bar"),
             Regex("/*"),
-        )))
+        ))
+        self.assertTrue(result)
         self.assertEqual("https", result['protocol'])
         self.assertEqual("somehost", result['host'])
         self.assertEqual("1", result['foo'])
         self.assertEqual("2", result['bar'])
 
-    # noinspection PyUnresolvedReferences
     def test_string_argresult(self):
         path = "https://somehost/foo=1/bar=2/"
 
-        self.assertTrue(result := match(path, String(
+        result = match(path, String(
             Capture(OneOf("https", "http"), name="protocol"),
             "://",
             Capture(Regex("[a-zA-Z_-]+"), name="host"),
@@ -274,7 +278,29 @@ class BasicUseCases(unittest.TestCase):
             "bar=",
             Capture(Regex("[^=/]+"), name="bar"),
             Regex("/*"),
-        ), argresult=True))
+        ))
+        self.assertTrue(result)
+        self.assertEqual("https", result.protocol)
+        self.assertEqual("somehost", result.host)
+        self.assertEqual("1", result.foo)
+        self.assertEqual("2", result.bar)
+
+    def test_string_with_at_capture(self):
+        path = "https://somehost/foo=1/bar=2/"
+
+        result = match(path, String(
+            'protocol' @ OneOf("https", "http"),
+            "://",
+            'host' @ Regex("[a-zA-Z_-]+"),
+            Regex("/+"),
+            "foo=",
+            'foo' @ Regex("[^=/]+"),
+            "/",
+            "bar=",
+            'bar' @ Regex("[^=/]+"),
+            Regex("/*"),
+        ))
+        self.assertTrue(result)
         self.assertEqual("https", result.protocol)
         self.assertEqual("somehost", result.host)
         self.assertEqual("1", result.foo)
@@ -295,7 +321,7 @@ class BasicUseCases(unittest.TestCase):
         }
 
         # noinspection PyUnresolvedReferences
-        self.assertTrue(result := match(request, Object(
+        result = match(request, Object(
             api_version="v1",
             job=Object(
                 run_at=Transformed(datetime.fromisoformat, 'time' @ _),
@@ -303,7 +329,8 @@ class BasicUseCases(unittest.TestCase):
                 Object(command='command' @ InstanceOf(str)),
                 Object(spawn='container' @ InstanceOf(str)),
             )
-        )))
+        ))
+        self.assertTrue(result)
         self.assertFalse('container' in result)
         self.assertEqual(result['time'], datetime(2020, 8, 27, 14, 9, 30))
         self.assertEqual(result['command'], "echo 'booya'")
@@ -331,15 +358,42 @@ class BasicUseCases(unittest.TestCase):
                 dct[k] = v
             return dct
 
-        self.assertTrue(result := match({'a': 1, 'b': 2, 'c': 3}, Object(a=_ >> 'x', b=_ >> 'y', c=_ >> 'z')))
+        result = match({'a': 1, 'b': 2, 'c': 3}, Object(a=_ >> 'x', b=_ >> 'y', c=_ >> 'z'))
+        self.assertTrue(result)
         r = f(**result)
         self.assertEqual({'x': 1, 'y': 2, 'z': 3}, r)
 
     def test_maybe(self):
         pattern = Object(a=...) & Maybe(Object(b=_ >> 'capture'))
 
-        self.assertTrue(result := match({'a': 3}, pattern))
+        result = match({'a': 3}, pattern)
+        self.assertTrue(result)
         self.assertFalse('capture' in result)
-        self.assertTrue(result := match({'a': 3, 'b': 5}, pattern))
+        result = match({'a': 3, 'b': 5}, pattern)
+        self.assertTrue(result)
         self.assertTrue('capture' in result)
         self.assertEqual(5, result['capture'])
+
+    def test_dataclasses(self):
+        @dataclass
+        class User:
+            first_name: str
+            last_name: str
+
+        @dataclass
+        class LoggedInUser(User):
+            display_name: str
+
+        result = match(User("Jane", "Doe"), User("Jane", _ >> 'last_name'))
+        self.assertTrue(result)
+        self.assertEqual("Doe", result['last_name'])
+
+        result = match(LoggedInUser("John", "Doe", "johndoe"), User(_, _ >> 'last_name'))
+        self.assertTrue(result)
+        self.assertEqual("Doe", result['last_name'])
+
+    def test_bind(self):
+        result = {}
+        if match({'a': 1, 'b': 2}, {'a': _ >> 'foo'}).bind(result):
+            self.assertEqual(1, result['foo'])
+        self.assertEqual({'foo': 1}, result)
